/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy.mock;

import static java.util.Objects.*;

import java.io.IOException;
import java.io.InputStream;
import java.util.Set;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.tuple.Pair;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAttachmentMetadata;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAttachments;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcBayernIdMessageMetadata;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import io.grpc.stub.CallStreamObserver;
import lombok.extern.log4j.Log4j2;

@Log4j2
class MockChunkedDataStreamer {

	private Set<Pair<GrpcAttachmentMetadata, InputStream>> attachments;
	private final int chunkSize;
	private boolean isFinished = false;
	private GrpcBayernIdMessageMetadata messageMetadata;

	public MockChunkedDataStreamer(GrpcBayernIdMessageMetadata metadata, Set<Pair<GrpcAttachmentMetadata, InputStream>> attachments, int chunkSize) {
		this.messageMetadata = metadata;
		this.attachments = attachments;
		this.chunkSize = chunkSize;
	}

	public synchronized void sendChunkedTo(CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		if (isFinished) {
			return;
		}
		sendMessageMetadata(streamObserver);
		attachments.forEach(attachment -> {
			sendAttachmentMetadata(attachment.getLeft(), streamObserver);
			sendContentChunked(attachment.getRight(), streamObserver);
		});
		handleFileEndReached(streamObserver);
	}

	void sendContentChunked(InputStream contentStream, CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		var inProgerss = true;
		do {
			int size = sendNextChunk(contentStream, streamObserver);
			if (size < chunkSize) {
				sendNextChunk(contentStream, streamObserver);
				inProgerss = false;
			}
		} while (inProgerss && streamObserver.isReady());
		IOUtils.closeQuietly(contentStream);
	}

	void sendMessageMetadata(CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		if (nonNull(messageMetadata)) {
			streamObserver.onNext(buildRequestWithMetadata(messageMetadata));
			messageMetadata = null;
		}
	}

	GrpcSendBayernIdMessageRequest buildRequestWithMetadata(GrpcBayernIdMessageMetadata metadata) {
		return GrpcSendBayernIdMessageRequest.newBuilder().setMessageMetadata(metadata).build();
	}

	void sendAttachmentMetadata(GrpcAttachmentMetadata metadata, CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		streamObserver.onNext(buildRequestWithAttachmentMetadata(metadata));
	}

	GrpcSendBayernIdMessageRequest buildRequestWithAttachmentMetadata(GrpcAttachmentMetadata attachmentMetadata) {
		return GrpcSendBayernIdMessageRequest.newBuilder()
				.setAttachments(GrpcAttachments.newBuilder().setAttachmentMetadata(attachmentMetadata).build()).build();
	}

	private int sendNextChunk(InputStream contentStream, CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		try {
			var content = contentStream.readNBytes(chunkSize);
			var size = content.length;
			if (size > 0) {
				streamObserver.onNext(buildRequestWithContent(content));
			}
			return size;
		} catch (IOException e) {
			throw new TechnicalException("Error on sending a single chunk", e);
		}
	}

	GrpcSendBayernIdMessageRequest buildRequestWithContent(byte[] bytes) {
		return GrpcSendBayernIdMessageRequest.newBuilder().setAttachments(GrpcAttachments.newBuilder().setContent(ByteString.copyFrom(bytes)).build())
				.build();
	}

	synchronized void handleFileEndReached(CallStreamObserver<GrpcSendBayernIdMessageRequest> streamObserver) {
		isFinished = true;
		streamObserver.onCompleted();
	}

}
